import sys

from threading import Thread
from threading import RLock
from fastrlock.rlock import FastRLock as FLock

# Benchmark functions:

cython_code = []

def lock_unlock(l):
    l.acquire()
    l.release()
    l.acquire()
    l.release()
    l.acquire()
    l.release()
    l.acquire()
    l.release()
    l.acquire()
    l.release()


cython_code.append("""
def lock_unlock(lock):
    lock_fastrlock(lock, -1, True)
    unlock_fastrlock(lock)
    lock_fastrlock(lock, -1, True)
    unlock_fastrlock(lock)
    lock_fastrlock(lock, -1, True)
    unlock_fastrlock(lock)
    lock_fastrlock(lock, -1, True)
    unlock_fastrlock(lock)
    lock_fastrlock(lock, -1, True)
    unlock_fastrlock(lock)
""")


def reentrant_lock_unlock(l):
    l.acquire()
    l.acquire()
    l.acquire()
    l.acquire()
    l.acquire()
    l.release()
    l.release()
    l.release()
    l.release()
    l.release()


cython_code.append("""
def reentrant_lock_unlock(lock):
    lock_fastrlock(lock, -1, True)
    lock_fastrlock(lock, -1, True)
    lock_fastrlock(lock, -1, True)
    lock_fastrlock(lock, -1, True)
    lock_fastrlock(lock, -1, True)
    unlock_fastrlock(lock)
    unlock_fastrlock(lock)
    unlock_fastrlock(lock)
    unlock_fastrlock(lock)
    unlock_fastrlock(lock)
""")


def mixed_lock_unlock(l):
    l.acquire()
    l.release()
    l.acquire()
    l.acquire()
    l.release()
    l.acquire()
    l.release()
    l.acquire()
    l.release()
    l.release()


cython_code.append("""
def mixed_lock_unlock(lock):
    lock_fastrlock(lock, -1, True)
    unlock_fastrlock(lock)
    lock_fastrlock(lock, -1, True)
    lock_fastrlock(lock, -1, True)
    unlock_fastrlock(lock)
    lock_fastrlock(lock, -1, True)
    unlock_fastrlock(lock)
    lock_fastrlock(lock, -1, True)
    unlock_fastrlock(lock)
    unlock_fastrlock(lock)
""")


def context_manager(l):
    with l: pass
    with l:
        with l:
            with l: pass
            with l: pass
        with l:
            with l: pass
        with l:
            with l: pass
            with l: pass
    with l: pass
    with l:
        with l:
            with l: pass
            with l: pass
            with l:
                with l: pass
    with l: pass


def lock_unlock_nonblocking(l):
    if l.acquire(False):
        l.release()
    if l.acquire(False):
        l.release()
    if l.acquire(False):
        l.release()
    if l.acquire(False):
        l.release()
    if l.acquire(False):
        l.release()


cython_code.append("""
def lock_unlock_nonblocking(lock):
    if lock_fastrlock(lock, -1, False):
        unlock_fastrlock(lock)
    if lock_fastrlock(lock, -1, False):
        unlock_fastrlock(lock)
    if lock_fastrlock(lock, -1, False):
        unlock_fastrlock(lock)
    if lock_fastrlock(lock, -1, False):
        unlock_fastrlock(lock)
    if lock_fastrlock(lock, -1, False):
        unlock_fastrlock(lock)
""")


# End of benchmark functions


def threaded(l, test_func, tcount=10):
    threads = [ Thread(target=test_func, args=(l,)) for _ in range(tcount) ]
    for thread in threads:
        thread.start()
    for thread in threads:
        thread.join()


def run_benchmark(name, lock, version, functions, repeat_count, repeat_count_t):
    print('Testing %s (%s)' % (name, version))

    from timeit import Timer
    from functools import partial

    print("sequential (x%d):" % repeat_count)
    for function in functions:
        timer = Timer(partial(function, lock))
        print('%-25s: %9.2f msec' % (function.__name__, max(timer.repeat(repeat=4, number=repeat_count)) * 1000.0))

    print("threaded 10T (x%d):" % repeat_count_t)
    for function in functions:
        timer = Timer(partial(threaded, lock, function))
        print('%-25s: %9.2f msec' % (function.__name__, max(timer.repeat(repeat=4, number=repeat_count_t)) * 1000.0))


if sys.version_info < (3, 5):
    import imp

    def load_dynamic(name, module_path):
        return imp.load_dynamic(name, module_path)
else:
    import importlib.util as _importlib_util

    def load_dynamic(name, module_path):
        spec = _importlib_util.spec_from_file_location(name, module_path)
        module = _importlib_util.module_from_spec(spec)
        spec.loader.exec_module(module)
        return module


def main():
    functions = [
        lock_unlock,
        reentrant_lock_unlock,
        mixed_lock_unlock,
        lock_unlock_nonblocking,
        context_manager,
        ]

    import fastrlock

    import glob
    import os.path
    import re
    import tempfile

    repeat_count = 100000
    repeat_count_t = 1000

    rlock = (RLock(), "%d.%d.%d" % sys.version_info[:3])
    flock = (FLock(), fastrlock.__version__)

    locks = []
    args = sys.argv[1:]
    if 'rlock' in args:
        locks.append(rlock)
    if 'flock' in args:
        locks.append(flock)
    if not args:
        locks = [rlock, flock]

    for _ in range(args.count('quick')):
        repeat_count = max(10, repeat_count // 100)
        repeat_count_t = max(5, repeat_count_t // 10)

    for lock, version in locks:
        name = type(lock).__name__
        run_benchmark(name, lock, version, functions, repeat_count, repeat_count_t)

    if 'cython' in args or not args:
        lock, version = flock

        from Cython.Build.Cythonize import cython_compile, parse_args

        basepath = None
        try:
            with tempfile.NamedTemporaryFile(mode="w", suffix='.pyx') as f:
                code = '\n'.join(cython_code)
                cy_function_names = re.findall(r"def (\w+)\(", code)
                f.write("from fastrlock.rlock cimport lock_fastrlock, unlock_fastrlock\n\n")
                f.write(code)
                f.flush()

                options, _ = parse_args(["", f.name, "-3", "--force", "--inplace"])
                cython_compile(f.name, options)

            basepath = os.path.splitext(f.name)[0]
            for ext in [".*.so", ".*.pyd", ".so", ".pyd"]:
                so_paths = glob.glob(basepath + ext)
                if so_paths:
                    so_path = so_paths[0]
                    break
            else:
                print("Failed to find Cython compiled module")
                sys.exit(1)

            module = load_dynamic(os.path.basename(basepath), so_path)

            cy_functions = [getattr(module, name) for name in cy_function_names]

            run_benchmark("Cython interface of %s" % type(lock).__name__, lock, version,
                          cy_functions, repeat_count, repeat_count_t)

        finally:
            if basepath:
                files = glob.glob(basepath + ".*")  # .c, .so / .pyd
                if len(files) > 3:
                    print("Found too many artefacts, not deleting temporary files")
                else:
                    for filename in files:
                        os.unlink(filename)


if __name__ == '__main__':
    main()




def suite():
    import unittest
    import doctest
    import os
    test_dir = os.path.abspath(os.path.dirname(__file__))

    tests = []
    for filename in os.listdir(test_dir):
        if filename.endswith('.py') and not filename.startswith('_'):
            tests.append(unittest.defaultTestLoader.loadTestsFromName(__name__ + '.' + filename[:-3]))

    suite = unittest.TestSuite(tests) if len(tests) != 1 else tests[0]
    suite.addTest(doctest.DocFileSuite('../../README.rst'))
    return suite


if __name__ == '__main__':
    import unittest
    unittest.main()




import sys
import os
import re
from setuptools import setup, Extension

with open('fastrlock/__init__.py') as f:
    VERSION = re.search(
        r'^__version__ \s* = \s* ["\'] ([0-9ab.]+) ["\']',
        f.read(), re.MULTILINE|re.VERBOSE).group(1)

BASEDIR = os.path.dirname(__file__)
PKGNAME = 'fastrlock'
PKGDIR = os.path.join(BASEDIR, PKGNAME)
MODULES = [filename[:-4] for filename in os.listdir(PKGDIR)
           if filename.endswith('.pyx')]


def has_option(name):
    if name in sys.argv[1:]:
        sys.argv.remove(name)
        return True
    return False


ext_args = {
    'define_macros': [('CYTHON_CLINE_IN_TRACEBACK', '1')],
}
if has_option('--without-assert'):
    ext_args['define_macros'].append(('CYTHON_WITHOUT_ASSERTIONS', None))


use_cython = has_option('--with-cython')
if not use_cython:
    if not all(os.path.isfile(os.path.join(PKGDIR, module_name+'.c'))
               for module_name in MODULES):
        print("NOTE: generated sources not available, need Cython to build")
        use_cython = True


cythonize = None
if use_cython:
    try:
        from Cython.Build import cythonize
        import Cython.Compiler.Version
        print("building with Cython " + Cython.Compiler.Version.version)
        source_extension = ".pyx"
    except ImportError:
        print("WARNING: trying to build with Cython, but it is not installed")
        cythonize = None
        source_extension = ".c"
else:
    print("building without Cython")
    source_extension = ".c"


ext_modules = [
    Extension(
        '%s.%s' % (PKGNAME, module_name),
        sources=[os.path.join(PKGNAME, module_name+source_extension)],
        **ext_args)
    for module_name in MODULES]


if cythonize is not None:
    ext_modules = cythonize(ext_modules)


def read_file(filename):
    f = open(os.path.join(BASEDIR, filename))
    try:
        return f.read()
    finally:
        f.close()


long_description = '\n\n'.join([
    read_file(text_file)
    for text_file in ['README.rst', 'CHANGES.rst']])


setup(
    name="fastrlock",
    version=VERSION,
    author="Stefan Behnel",
    author_email="stefan_ml@behnel.de",
    url="https://github.com/scoder/fastrlock",
    license='MIT style',
    description="Fast, re-entrant optimistic lock implemented in Cython",
    long_description=long_description,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Cython',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
        'Topic :: Software Development',
    ],

    packages=[PKGNAME],
    package_data={PKGNAME: ['*.pxd', '*.pxi']},
    ext_modules=ext_modules,
    zip_safe=False,

    # support 'test' target if setuptools/distribute is available
    test_suite='fastrlock.tests.suite',
)



# -*- coding: utf-8 -*-

try:
    from thread import start_new_thread, get_ident
except ImportError:
    # Python 3?
    from _thread import start_new_thread, get_ident
import threading
import unittest
import time
import sys
import gc

from fastrlock import rlock

IS_PYTHON3 = sys.version_info[0] >= 3

try:
    _next = next
except NameError:
    def _next(o):
        return o.next()

unicode_type = type(IS_PYTHON3 and 'abc' or 'abc'.decode('ASCII'))


def _wait():
    # A crude wait/yield function not relying on synchronization primitives.
    time.sleep(0.01)


class Bunch(object):
    """
    A bunch of threads.
    """
    def __init__(self, f, n, wait_before_exit=False):
        """
        Construct a bunch of `n` threads running the same function `f`.
        If `wait_before_exit` is True, the threads won't terminate until
        do_finish() is called.
        """
        self.f = f
        self.n = n
        self.started = []
        self.finished = []
        self._can_exit = not wait_before_exit
        def task():
            tid = get_ident()
            self.started.append(tid)
            try:
                f()
            finally:
                self.finished.append(tid)
                while not self._can_exit:
                    _wait()
        for i in range(n):
            start_new_thread(task, ())

    def wait_for_started(self):
        while len(self.started) < self.n:
            _wait()

    def wait_for_finished(self):
        while len(self.finished) < self.n:
            _wait()

    def do_finish(self):
        self._can_exit = True


################################################################################
# tests for the FastRLock implementation

class TestFastRLock(unittest.TestCase):
    """Copied from CPython's test.lock_tests module
    """
    locktype = rlock.FastRLock
    Bunch = Bunch

    def tearDown(self):
        gc.collect()

    # the locking tests

    """
    Tests for both recursive and non-recursive locks.
    """

    def test_constructor(self):
        lock = self.locktype()
        del lock

    def test_acquire_destroy(self):
        lock = self.locktype()
        lock.acquire()
        del lock

    def test_acquire_release(self):
        lock = self.locktype()
        lock.acquire()
        lock.release()
        del lock

    def test_try_acquire(self):
        lock = self.locktype()
        self.assertTrue(lock.acquire(False))
        lock.release()

    def test_try_acquire_contended(self):
        lock = self.locktype()
        lock.acquire()
        result = []
        def f():
            result.append(lock.acquire(False))
        self.Bunch(f, 1).wait_for_finished()
        self.assertFalse(result[0])
        lock.release()

    def test_acquire_contended(self):
        lock = self.locktype()
        lock.acquire()
        N = 5
        def f():
            lock.acquire()
            lock.release()

        b = self.Bunch(f, N)
        b.wait_for_started()
        _wait()
        self.assertEqual(len(b.finished), 0)
        lock.release()
        b.wait_for_finished()
        self.assertEqual(len(b.finished), N)

    def test_with(self):
        lock = self.locktype()
        def f():
            lock.acquire()
            lock.release()
        def _with(err=None):
            with lock:
                if err is not None:
                    raise err
        _with()
        # Check the lock is unacquired
        self.Bunch(f, 1).wait_for_finished()
        self.assertRaises(TypeError, _with, TypeError)
        # Check the lock is unacquired
        self.Bunch(f, 1).wait_for_finished()

    def test_thread_leak(self):
        # The lock shouldn't leak a Thread instance when used from a foreign
        # (non-threading) thread.
        lock = self.locktype()
        def f():
            lock.acquire()
            lock.release()
        n = len(threading.enumerate())
        # We run many threads in the hope that existing threads ids won't
        # be recycled.
        self.Bunch(f, 15).wait_for_finished()
        self.assertEqual(n, len(threading.enumerate()))

    """
    Tests for non-recursive, weak locks
    (which can be acquired and released from different threads).
    """

    def _test_reacquire_nonrecursive(self):
        # Lock needs to be released before re-acquiring.
        lock = self.locktype()
        phase = []
        def f():
            lock.acquire()
            phase.append(None)
            lock.acquire()
            phase.append(None)
        start_new_thread(f, ())
        while len(phase) == 0:
            _wait()
        _wait()
        self.assertEqual(len(phase), 1)
        lock.release()
        while len(phase) == 1:
            _wait()
        self.assertEqual(len(phase), 2)

    def test_different_thread_release_succeeds(self):
        # Lock can be released from a different thread.
        lock = self.locktype()
        lock.acquire()
        def f():
            lock.release()
        b = self.Bunch(f, 1)
        b.wait_for_finished()
        lock.acquire()
        lock.release()

    """
    Tests for recursive locks.
    """

    def test_reacquire(self):
        lock = self.locktype()
        lock.acquire()
        lock.acquire()
        lock.release()
        lock.acquire()
        lock.release()
        lock.release()

    def test_release_unacquired(self):
        # Cannot release an unacquired lock
        lock = self.locktype()
        self.assertRaises(RuntimeError, lock.release)
        lock.acquire()
        lock.acquire()
        lock.release()
        lock.acquire()
        lock.release()
        lock.release()
        self.assertRaises(RuntimeError, lock.release)

    def test_different_thread_release_fails(self):
        # Cannot release from a different thread
        lock = self.locktype()
        def f():
            lock.acquire()
        b = self.Bunch(f, 1, True)
        try:
            self.assertRaises(RuntimeError, lock.release)
        finally:
            b.do_finish()

    def test__is_owned(self):
        lock = self.locktype()
        self.assertFalse(lock._is_owned())
        lock.acquire()
        self.assertTrue(lock._is_owned())
        lock.acquire()
        self.assertTrue(lock._is_owned())
        result = []
        def f():
            result.append(lock._is_owned())
        self.Bunch(f, 1).wait_for_finished()
        self.assertFalse(result[0])
        lock.release()
        self.assertTrue(lock._is_owned())
        lock.release()
        self.assertFalse(lock._is_owned())


if __name__ == '__main__':
    import doctest
    suite = unittest.TestSuite()
    suite.addTest(doctest.DocTestSuite(rlock))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromName('__main__'))
    suite.addTest(doctest.DocFileSuite('../../README.rst'))
    runner = unittest.TextTestRunner(verbosity=2)
    if not runner.run(suite).wasSuccessful():
        sys.exit(1)
